package binsec_codex

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Binsec_codex_libSource

Sourcemodule Analyze : sig ... end
module Arch_settings : sig ... end

This module signature describes two rather distinct components:

Sourcemodule Benchmark : sig ... end
Sourcemodule Binsec2syntax_tree : sig ... end

This file provides the decomposition from specific Dba.Expr.t to the language-agnostic Syntax_tree.t, and checks that the index used to identify expressions are correct (internally, using a single function as the source of truth for this decomposition).

Sourcemodule Cfg_analysis : sig ... end

A graph of (address, call_stack) pair that can be translated to regular expressions. It can be used to store a graph of instructions or a graph of basic blocks.

Sourcemodule Codex_logger : sig ... end
Sourcemodule Codex_options : sig ... end
Sourcemodule Dba2Codex : sig ... end
Sourcemodule Dhunk_analysis : sig ... end
Sourcemodule Heap_typing : sig ... end
Sourcemodule Hooks : sig ... end
Sourcemodule Html : sig ... end
Sourcemodule Interval2symbol : sig ... end
Sourcemodule Record_cfg : sig ... end
Sourcemodule X86_arch : sig ... end