package binsec_codex

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Create.MakeSource

Parameters

module Reg : Arch_settings.Registers with module Domain = Domain

Signature

Sourcemodule State : StateS with module Domain = Domain

Analyse a DBA instruction starting from a state. idx is the position of the expression, and is used to precisely track the location in the source code.

Analyse a DBA instruction starting from a state. idx is the index of the DBA instruction in the dhunk, and is used to precisely track the location in the source code.