package binsec_codex

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Binsec_codex_lib.Dhunk_analysisSource

Sourcemodule V : sig ... end
Sourcemodule E : sig ... end
Sourcemodule G : sig ... end
Sourcemodule Reduce : sig ... end
Sourcemodule Wto_alg : sig ... end

For each outer jump instruction in a dhunk, return the path expression of that DBA instruction in the dhunk, and optionnally the target address, if it statically known.